"""pcassie: PCA reduction of exoatmosphere spectra"""
__version__ = "0.1.0"

from .calibration import * 
from .ccf_tests import * 
from .ccf import * 
from .pca_subtraction import * 
from .pipeline import * 
from .plot_functions import * 
from .utility_functions import * 




import numpy as np
import matplotlib.pyplot as plt

from pcassie.pca_subtraction import *

plt.rcParams.update({'axes.linewidth' : 1.5,
                     'ytick.major.width' : 1.5,
                     'ytick.minor.width' : 1.5,
                     'xtick.major.width' : 1.5,
                     'xtick.minor.width' : 1.5,
                     'xtick.labelsize': 12, 
                     'ytick.labelsize': 12,
                     'axes.labelsize': 18,
                     'axes.labelpad' : 5,
                     'axes.titlesize' : 22,
                     'axes.titlepad' : 10,
                     'font.family': 'Serif'
                    })

def plot_spectral_square(spectra_array, wave, title=None, x_label=None, y_label=None, cbar_label=None):
    """Plots a 2d "spectra_array" grid against a 1d "wave" array. Valuable to show a grid od spectral observations, 
    2d CCF arrays, etc."""
    n_spec = spectra_array.shape[0]

    # Create 2D grid for wave and index
    wave_edges = np.concatenate([
        [wave[0] - (wave[1] - wave[0]) / 2],
        (wave[1:] + wave[:-1]) / 2,
        [wave[-1] + (wave[-1] - wave[-2]) / 2]
    ])
    idx = np.arange(n_spec + 1)

    plt.figure(figsize=(10, 5))
    mesh = plt.pcolormesh(wave_edges, idx, spectra_array, shading='auto',
                          cmap='viridis',
                          vmin=np.percentile(spectra_array, 1),
                          vmax=np.percentile(spectra_array, 99))
    plt.colorbar(mesh, label=cbar_label or 'Flux')
    plt.xlabel(x_label or 'Wavelength')
    plt.ylabel(y_label or 'Spectrum Index')
    plt.title(title or 'Spectral Square Plot')
    plt.tight_layout()
    plt.show()

def plot_preprocess(flux, wave):
    """Plots the spectral grid at various stages of normalizing, 
    median subtracting, and standard deviation dividing."""
    plot_spectral_square(flux, wave, title="Base Spectra")

    # Normalize by the median of this spectrum
    norm_flux = flux / np.median(flux)

    plot_spectral_square(flux, wave, title="Normalized Spectra")

    # Compute the median at each wavelength (column)
    median_flux = np.median(flux, axis=0)

    # Subtract the median from each spectrum
    median_subtracted_flux = norm_flux - median_flux  # shape: (num_spectra, num_wavelengths)

    plot_spectral_square(median_subtracted_flux, wave, title="Median Subtracted Spectra")

    # Compute the standard deviation for each spectrum (row)
    row_std = np.std(median_subtracted_flux, axis=1, keepdims=True)  # shape: (num_spectra, 1)

    # Divide each row by its own standard deviation
    row_std_divided_flux = median_subtracted_flux / row_std  # shape: (num_spectra, num_wavelengths)

    plot_spectral_square(row_std_divided_flux, wave, title="Standard Deviation Divided Spectra")

    return row_std_divided_flux

def plot_covariance(tdm_covariance, wdm_covariance):
    """Plots the covariance grid of the PCA analysis in the time and wavelegnth domains."""
    plt.imshow(tdm_covariance, cmap='viridis', aspect='auto')
    plt.colorbar(label='Covariance')
    plt.title("TDM Covariance Matrix")
    plt.xlabel("Wavelength Index")
    plt.ylabel("Wavelength Index")
    plt.show()

    plt.imshow(wdm_covariance, cmap='viridis', aspect='auto')
    plt.colorbar(label='Covariance')    
    plt.title("WDM Covariance Matrix")
    plt.xlabel("Spectrum Index")
    plt.ylabel("Spectrum Index")
    plt.show()

def plot_eigenvectors(eigenvectors, title=None):
    """Plots the first five eigenvectors."""
    _, axes = plt.subplots(5, 1, figsize=(10, 12), sharex=True)
    for i in range(5):
        axes[i].plot(eigenvectors[:, i], label=f'Eigenvector {i+1}')
        axes[i].set_ylabel('Value')
        axes[i].legend(loc='upper right')
        if i == 0 and title:
            axes[i].set_title(title)
    axes[-1].set_xlabel('Index')
    plt.tight_layout()
    plt.show()

def plot_explained_variance(eigenvalues, title=None):
    """Plot the explained variance from eigenvalues."""
    explained_var = explained_variance(eigenvalues)
    plt.figure(figsize=(10, 6))
    plt.plot(explained_var, marker='o', linestyle='-', color='b')
    plt.title('Explained Variance by Eigenvalues' if title is None else title)
    plt.xlabel('Eigenvalue Index')
    plt.ylabel('Explained Variance')
    plt.yscale("log", base=10)
    plt.grid()
    plt.show()

def plot_reconstructed_spectra(original, reconstructed, wave, title=None):
    """Plots the original and post-PCA spectra."""
    plt.figure(figsize=(10, 6))
    plt.plot(wave, original[0], label='Original Spectrum', alpha=0.5)
    plt.plot(wave, reconstructed[0], label='Reconstructed Spectrum', linestyle='--')
    plt.xlabel('Wavelength')
    plt.ylabel('Flux')
    if title:
        plt.title(title)
    else:
        plt.title('Original vs Reconstructed Spectrum')
    plt.legend()
    plt.show()

def plot_pca_subtraction(spectra, wave, start_wav, end_wav, first_comps=0, last_comps=0, preprocess=False):
    """Runs PCA subtraction and plots the results."""
    if preprocess:
        print("Preprocessing spectra...")
        spectra = plot_preprocess(spectra, wave)
    else:
        print("Skipping preprocessing...")

    start_idx, end_idx = convert_range_to_indices(wave, start_wav, end_wav)

    tdm_df = pd.DataFrame(spectra[:, start_idx:end_idx].T)
    wdm_df = pd.DataFrame(spectra[:, start_idx:end_idx])

    tdm_covariance = tdm_df.cov().values
    wdm_covariance = wdm_df.cov().values

    eval_tdm, evec_tdm  = compute_eigenvalues_and_vectors(tdm_covariance)
    eval_wdm, evec_wdm = compute_eigenvalues_and_vectors(wdm_covariance)

    # Remove components from TDM and WDM
    tdm_reconstructed = remove_components(spectra[:, start_idx:end_idx].T, evec_tdm, first_comps, last_comps)
    wdm_reconstructed = remove_components(spectra[:, start_idx:end_idx], evec_wdm, first_comps, last_comps)

    plot_covariance(tdm_covariance, wdm_covariance)

    plot_eigenvectors(evec_tdm, title="TDM Eigenvectors")
    plot_eigenvectors(evec_wdm, title="WDM Eigenvectors")

    plot_explained_variance(eval_tdm, title="TDM Explained Variance")
    plot_explained_variance(eval_wdm, title="WDM Explained Variance")

    plot_reconstructed_spectra(spectra[:, start_idx:end_idx], tdm_reconstructed.T, wave[start_idx:end_idx], title="TDM Reconstructed Spectrum")
    plot_reconstructed_spectra(spectra[:, start_idx:end_idx], wdm_reconstructed, wave[start_idx:end_idx], title="WDM Reconstructed Spectrum")


### CCF Plot Functions 

def plot_intransit_ccfs(planet_frame_vgrid, in_transit, mean_subtracted=False):
    """Plots the velocity vs. CCF value for all spectra taken at the time of transit, 
    as well as their co-added sum. mean_subtracted effectively toggles normalization."""
    plt.figure(figsize=(10, 6))

    if mean_subtracted:
        sum = np.zeros_like(planet_frame_vgrid)
        for i, ccf in enumerate(in_transit):
            ccf -= np.mean(ccf)  # Normalize each CCF by subtracting the mean
            sum += ccf
            plt.plot(np.array(planet_frame_vgrid) / 1000, ccf, label=f"Spectrum {i+1}")
        plt.plot(planet_frame_vgrid / 1000, sum, label="Mean Subtracted Sum", color='black', linewidth=2)
        plt.title("Mean-Subtracted In-transit CCFs")

    else: 
        sum = np.zeros_like(planet_frame_vgrid)
        for i, ccf in enumerate(in_transit):
            sum += ccf
            plt.plot(np.array(planet_frame_vgrid) / 1000, ccf, label=f"Spectrum {i+1}")
        plt.plot(planet_frame_vgrid / 1000, sum, label="Sum", color='black', linewidth=2)
        plt.title("In-transit CCFs")

    plt.xlabel(r"Velocity $[kms^{-1}]$")
    plt.ylabel("CCF co-added value")
    plt.legend(ncol=3, loc='lower right', fontsize='small')
    plt.grid()
    plt.show()

    return sum

## CCF TEST PLOT FUNCTIONS

def plot_welch_t_test(in_trail_vals, out_of_trail_vals, t_stat, p_value, bins=None): 
    """Plots a histogram of Welch's T-test values for the range of values associated 
    with the planet (in trail) vs. those outside (out trail)."""
    plt.figure(figsize=(10, 6))
    bins = bins or [0.15, 0.20, 0.25, 0.30, 0.35, 0.40, 0.45, 0.50] 

    plt.hist(out_of_trail_vals, bins=bins, label='Out-of-trail', color='white', histtype='step', edgecolor='blue', density=True)
    plt.hist(in_trail_vals, bins=bins, label='In-trail', color='white', histtype='step', edgecolor='orange', density=True)

    plt.axvline(np.mean(in_trail_vals), color='orange', linestyle='--', label='In-trail mean')
    plt.axvline(np.mean(out_of_trail_vals), color='blue', linestyle='--', label='Out-of-trail mean')

    plt.title(f"Welch’s t-test\nT = {t_stat:.2f}, p = {p_value:.2e}")
    plt.xlabel("CCF Value")
    plt.ylabel("Density")
    plt.legend()
    plt.tight_layout()
    plt.show()



import numpy as np

# Functions to support normalization of CRIRES+ data

def normalize_zero_one(spectrum):
    median = np.nanmedian(spectrum)
    std = np.nanstd(spectrum)

    clean = spectrum.copy()

    outliers = np.abs(clean - median) > 3 * std
    clean[outliers] = median

    zero_to_one = (clean - np.min(clean)) / (np.max(clean) - np.min(clean))

    return zero_to_one


def split_divide_by_median(wave, flux, m):
    """
    Splits the spectrum into sections based on gaps in the wavelength array,
    normalizes each section by dividing by its median, and returns the normalized flux,
    the gap indices, and the segment index pairs (start, end).
    """
    valid_wave = wave[~np.isnan(wave)]
    gap_threshold = m * np.median(np.diff(valid_wave))
    gaps = np.where(np.diff(wave) > gap_threshold)[0]

    section_edges = np.concatenate(([0], gaps + 1, [len(wave)]))

    norm_flux = np.full_like(flux, np.nan)
    segment_indices = []

    for i in range(len(section_edges) - 1):
        start, end = section_edges[i], section_edges[i + 1]
        segment_indices.append((start, end))

        section = flux[start:end]
        valid_mask = ~np.isnan(section)
        if not np.any(valid_mask):
            continue

        median = np.nanmedian(section[valid_mask])
        std = np.nanstd(section[valid_mask])

        section_clean = section.copy()
        outliers = np.abs(section_clean - median) > 3 * std
        section_clean[outliers & valid_mask] = median

        section_norm = section_clean / median
        norm_flux[start:end] = section_norm

    return norm_flux, segment_indices


def split_detectors(wave, flux, m=5):
    normalized_flux = []

    for ii in range(len(flux)):
        single_flux = flux[ii, :]
        single_wave = wave[ii, :]
        
        single_norm_flux, segment_indices = split_divide_by_median(single_wave, single_flux, m)
        
        normalized_flux.append(single_norm_flux)

    normalized_flux_array = np.array(normalized_flux)

    return normalized_flux_array, segment_indices 


def mask_gap_edges(wave, gaps, n):
    """
    Returns a boolean mask that is True for points to be masked:
    - The first n points after each gap
    - The last n points before each gap
    - Optionally, the first n and last n points of the array
    """
    mask = np.zeros_like(wave, dtype=bool)
    for g in gaps:
        # Mask last n points before the gap (ending at g)
        start = max(0, g - n + 1)
        mask[start:g+1] = True
        # Mask first n points after the gap (starting at g+1)
        end = min(len(wave), g + 1 + n)
        mask[g+1:end] = True
    # Optionally, mask the first n and last n points of the array
    mask[:n] = True
    mask[-n:] = True
    return mask


def debug_print(verbose, *args, **kwargs):
    """General function to allow for toggleable print statements."""
    if verbose:
        print(*args, **kwargs)



import numpy as np

from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.time import Time
from numba import njit
from scipy.interpolate import interp1d

from pcassie.utility_functions import debug_print

@njit
def doppler_shift(wave_arr, velocity):
    """
    Doppler shift the wavelength array by a velocity (in m/s).
    
    Parameters
    ----------
    wave_arr: array
        1d wavelengths array (in any units, e.g., nm or µm)
    velocity: float
        Velocity in m/s

    Returns
    -------
    array 
        Doppler-shifted wavelengths
    """
    C = 299792458.0  # Speed of light in m/s
    return wave_arr * (1 + velocity / C)


def ccf(all_pca, all_wave, v_shift_range, sim_wave, sim_flux, speed=True, verbose=False):
    """Loops through all detectors and spectra to 
    compute the cross-correlation function of each spectrum 
    with the simulated spectrum.
    
    Parameters
    ----------
    all_pca: list
        PCA removed spectra with dimensions (n detectors x n spectra x wavelength grid).
    all_wave: list
        Wavelength arrays (n detectors x wavelength grid) for the corresponding PCA subtracted spectra. 
    v_shift_range: array
        1d array of velocity shifts to sample. Should be structured such that 
        the array has a step of 1 km/s, e.g. np.linspace(-100_000, 100_000, 201) with units of meters.
    sim_wave: array
        Simulated wave array.
    sim_flux: array
        Simulated flux array. 
    speed: boolean
        If True, uses numpy for interpolation (much faster). If False, uses scipy's interp1d (more accurate). 
        I reccomend using numpy for PCA sampling and scipy for science data.
        
    Returns
    -------
    array
        2d cross-correlation function array (n spectra x v_shift)"""
    sort_idx = np.argsort(sim_wave)
    sorted_wave = sim_wave[sort_idx]
    sorted_flux = sim_flux[sort_idx]

    stacked_segment_xcorr = []
    
    debug_print(verbose, "sorted wave, flux")
    for detector_spectra, detector_wavs in zip(all_pca, all_wave):
        detector_xcorr = []

        debug_print(verbose, f"cycling through {len(detector_spectra)} spectra")
        for single_data_spectra in detector_spectra:
            norm_xcorr_arr = []
            debug_print(verbose, f"single_data_spectra: {len(single_data_spectra)}")
            for v_shift in v_shift_range:
                # Doppler shift the template wavelength
                shifted_wave = doppler_shift(sorted_wave, v_shift)

                # Interpolate shifted flux onto segment wavelength grid
                if speed:   
                    # np.interp is faster but less accurate for v < 0; good for sampling the PCA space
                    shifted_flux = np.interp(detector_wavs, shifted_wave, sorted_flux)

                else:
                    # scipy's interp1d is slowe but more accurate for v < 0; good for getting science values 
                    interp_func = interp1d(shifted_wave, sorted_flux, bounds_error=False, fill_value=0.0)
                    shifted_flux = interp_func(detector_wavs)

                shifted_flux = (shifted_flux - np.mean(shifted_flux)) #/ np.std(shifted_flux)
                single_data_spectra = (single_data_spectra - np.mean(single_data_spectra)) #/ np.std(single_data_spectra)

                # Cross-correlate (dot product)
                xcorr = np.dot(shifted_flux, single_data_spectra)
                denom = np.sqrt(np.sum(shifted_flux**2) * np.sum(single_data_spectra**2))
                if denom == 0:
                    norm_xcorr = 0
                else:
                    norm_xcorr = xcorr / denom

                norm_xcorr_arr.append(norm_xcorr)

            norm_xcorr_arr = np.array(norm_xcorr_arr)
            debug_print(verbose, f"norm_xcorr_arr shape: {norm_xcorr_arr.shape}")
            detector_xcorr.append(norm_xcorr_arr)

        detector_xcorr = np.array(detector_xcorr)
        debug_print(verbose, f"detector_xcorr shape: {detector_xcorr.shape}")
        stacked_segment_xcorr.append(detector_xcorr)

    return np.sum(np.array(stacked_segment_xcorr), axis=0)


# doppler shift correction functions
def orbital_phase(t, T_not, P_orb):
    """Calculates the orbital phase as 
    phi(t) = (t - T_not) / P_orb
    where phi(t) is phase as a function of time, t is time, 
    T_not is the mid-transit time, and P_orb is the orbital period.
    Please ensure units match."""
    return (t - T_not)/P_orb


def orbit_velocity(a, P_orb):
    """Calculates orbital velocity as
    v_orb = 2*pi*a / P_orb
    where a is the semi-major axis and 
    P_orb is the orbital period.
    Please ensure units match."""
    return 2 * np.pi * a / P_orb 


def rv_amplitude(a, P_orb, i):
    """Calculates the radial velocity amplitude as:
    Kp = v_orb * sin(i)
    where v_orb is the orbital velocity, i is the inclination,
    a is the semi-major axis, and P_orb is the orbital period. 
    See orbit_velocity for the calculation of v_orb.
    Please ensure units match."""
    v_orb = orbit_velocity(a, P_orb)
    return v_orb * np.sin(i)


def doppler_correction(a, P_orb, i, t, T_not, v_sys, v_bary, Kp=None, verbose=False):
    """
    a in au, P_orb in days, i in degrees
    t in MJD
    T_not in MJD (mid-transit time)
    v_sys in km/s
    v_bary in km/s
    Kp in m/s, if None, will compute from a, P_orb, i
    """
    a = a * 1.495979e11  # Convert au to meters
    P_orb = P_orb * 24 * 3600  # Convert days to seconds
    i = np.radians(i)  # Convert degrees to radians
    t = t * 24 * 3600  # Convert MJD to seconds
    T_not = T_not * 24 * 3600  # Convert MJD to seconds
    v_bary = v_bary * 1000  # Convert km/s to m/s   
    v_sys = v_sys * 1000  # Convert km/s to m/s

    if Kp is None:
        Kp = rv_amplitude(a, P_orb, i)
    phi = orbital_phase(t, T_not, P_orb)

    debug_print(verbose, f"Kp: {Kp} m/s, orbital phase: {phi}")

    return (Kp * np.sin(2*np.pi*phi)) + v_sys + v_bary


def compute_vbary_timeseries(ra_deg, dec_deg, times_utc, location):
    """
    Compute v_bary(t) for a target at (ra, dec) and a time array.
    
    Parameters:
        ra_deg (float): RA in degrees
        dec_deg (float): Dec in degrees
        times_utc (array-like): List or array of UTC times (ISO strings or float MJD)
        location (EarthLocation): Astropy EarthLocation (observatory)

    Returns:
        np.ndarray: Barycentric velocities (km/s) for each time
    """
    target = SkyCoord(ra=ra_deg * u.deg, dec=dec_deg * u.deg)
    times = Time(times_utc, format='mjd', scale='utc', location=location)

    barycorr = target.radial_velocity_correction(obstime=times)
    return barycorr.to(u.km/u.s).value


def doppler_correct_ccf(summed_ccf, v_shift_range, mjd_obs, ra, dec, location, a, P_orb, i, T_not, v_sys, Kp=None, verbose=False):
    """Corrects the full cross-correlation array for the Doppler shift according to
    Vp = Kp*sin[2*pi*phi(t)] + v_sys + v_bary
    where Vp is the velocity correction, v_sys is the systems radial velocity, and 
    v_bary is the barcentric correction. Refer to rv_amplitude and orbital_phase 
    for the definitions of Kp and phi(t)"""
    v_bary_timeseries = compute_vbary_timeseries(ra, dec, mjd_obs, location)
    debug_print(verbose, f"v_bary_timeseries: {v_bary_timeseries}")
    all_doppler_corrects = []

    for jj in range(len(mjd_obs)):

        if Kp is None:
            correction = doppler_correction(a=a, P_orb=P_orb, i=i, t=mjd_obs[jj], T_not=T_not, v_sys=v_sys, v_bary=v_bary_timeseries[jj])

        else:
            correction = doppler_correction(a=a, P_orb=P_orb, i=i, t=mjd_obs[jj], T_not=T_not, v_sys=v_sys, v_bary=v_bary_timeseries[jj], Kp=Kp)

        all_doppler_corrects.append(correction)

    # check for nans in doppler correction
    debug_print(verbose, f"Doppler correction contains NaNs: {np.any(np.isnan(all_doppler_corrects))}")
    debug_print(verbose, f"Doppler corrections: {all_doppler_corrects}")

    new_vel_grids = []

    for kk in range(len(summed_ccf)):
        new_vel_grid = v_shift_range + all_doppler_corrects[kk]
        #debug_print(verbose, f"start, end of new_vel_grid: {new_vel_grid[0]}, {new_vel_grid[-1]}")
        new_vel_grids.append(new_vel_grid)

    min_v, max_v = -50000, 50000
    common_v_grid = np.linspace(min_v, max_v, 101)  # Common velocity grid for cropping

    #check for nans in new_vel_grids
    debug_print(verbose, f"New velocity grids contain NaNs: {np.any([np.any(np.isnan(v)) for v in new_vel_grids])}")

    cropped_ccf = []

    for i in range(len(new_vel_grids)):
        v = new_vel_grids[i]
        ccf = summed_ccf[i]
            
        common_mask = (v >= min_v) & (v <= max_v)
        debug_print(verbose, f"common velocity grid shape: {common_v_grid.shape}, v shape: {v.shape}, ccf shape: {ccf.shape}, common_mask sum: {np.sum(common_mask)}")
        interp_ccf = np.interp(common_v_grid, v[common_mask], ccf[common_mask])
        cropped_ccf.append(interp_ccf)

    #check for nans in cropped_ccf
    debug_print(verbose, f"Cropped CCF contains NaNs: {np.any(np.isnan(cropped_ccf))}")

    return np.array(cropped_ccf), common_v_grid 


def remove_out_of_transit(transit_start_end, grid, mjd_obs):
    """remove spectra outside of ingress (transit start) and egress (trasnit end)."""
    transit_start, transit_end = transit_start_end
    transit_mask = (mjd_obs >= transit_start) & (mjd_obs <= transit_end)
    filtered_grid = [grid[i] for i in range(grid.shape[0]) if transit_mask[i]]

    return filtered_grid


def run_ccf_on_detector_segments(all_wave, 
                                 all_pca, v_shift_range, segment_indices, sim_wave, 
                                 sim_flux, mjd_obs, ra, dec, location, 
                                 a, P_orb, i, T_not, v_sys, transit_start_end, verbose=False): #sim_wave in um for now
    """Full pipeline to runn cross-correlation analysis on your full dataset (n detectors x n spectra x wavelength range)"""
    
    earth_frame_ccf = ccf(all_pca, all_wave, v_shift_range, sim_wave, sim_flux, verbose=verbose)

    # check for NaNs in earth_frame_ccf
    debug_print(verbose, f"Earth frame CCF contains NaNs: {np.any(np.isnan(earth_frame_ccf))}")
    debug_print(verbose, f"Earth frame CCF shape: {earth_frame_ccf.shape}")
    planet_frame_ccf, planet_frame_vgrid = doppler_correct_ccf(earth_frame_ccf, v_shift_range, mjd_obs, ra, dec, location, a, P_orb, i, T_not, v_sys, verbose=verbose)

    in_transit = remove_out_of_transit(
    transit_start_end, planet_frame_ccf, mjd_obs)

    return earth_frame_ccf, planet_frame_ccf, planet_frame_vgrid, in_transit


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pcassie'
copyright = '2025, Kenny T. Phan'
author = 'Kenny T. Phan'
release = '0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.napoleon", 
    "sphinx.ext.viewcode", 
    "nbsphinx", 
    "sphinx.ext.autodoc"]

templates_path = ['_templates']
exclude_patterns = ['.ipynb_checkpoints/*']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'bizstyle'
html_static_path = ['_static']

# additions
nbsphinx_prompt_width = 0 # no prompts in nbsphinx
master_doc = 'index'


import numpy as np

from pcassie.utility_functions import split_detectors, debug_print
from pcassie.pca_subtraction import pca_subtraction
from pcassie.ccf import run_ccf_on_detector_segments
from pcassie.ccf_tests import sn_map, welch_t_test, find_max_sn_in_expected_range

def pipeline(sim_wave, sim_flux, v_shift_range=np.linspace(-100_000, 100_000, 201), Kp_range=np.linspace(50_000, 150_000, 101), verbose=True, eighcalc='numba', **kwargs):
    """Runs principal component analysis and cross-correlation with simulated signal. Additionally outputs
    a signal to noise map and Welch's T-test values. It is more legibple to run ``results = pipeline(...)`` 
    and handle the outputs as indices of ``results``.
     
    Parameters
    ----------
    sim_wave: array
        1d wavelength grid for the simulated specrum. Must be the same shape as kwargs['wave'].
    sim_flux: array
        1d flux values for the simulated spectrum. Must be the same shape as kwargs['wave'].
    v_shift_range: array
        1d range of velocity values to sample in the cross-correlation function. Ideally, this is 
        in units of meters/second and orgamized with a 1 km/s step, e.g. the default ``np.linspace(-100_000, 100_000, 201)``.
        At this stage, a 1 km/s step is assumed to be equivalent to 1 pixel in the CCF grid.
    verbose: boolean
        Boolean statement to toggle print statements throughout the pipeline.
    **kwargs: 
        For ease of use, ``pipeline`` takes keyword arguments for the many inputs it requires. This is typically neater organized as:
        ``obs_args = {'wave', 'flux', 'mjd_obs', 'ra', 'dec' 'location'}``, ``planet_args = {'a', 'P_orb', 'i', 'T_not', 'v_sys'}``,
        ``params = {'transit_start_end', 'gap_size', 'remove_segments', 'first_components', 'last_components'}`` and called as: 
        ``pipeline(sim_wave, sim_flux, **obs_args, **planet_args, **params)``. Please refer to the tutorial for an example use. 
            wave: array
                2d wavelength array of dimensions n sepectra x wavegrid. wavegrid must be consistent across all spectrum, i.e. this is 
                effectively a 1d array repeated for the number of spectra. This will be changed to only input a 1d array in later versions.
            flux: array
                2d flux array of dimensions n spectra x wavegrid. n spectra should be sorted with respect to time, i.e. ``flux[0]`` is the spectra 
                observed at ``mjd_obs[0]``.
            mjd_obs: list or array
                List or 1d array of the MJD-OBS value in the specta's ``.fits`` files. In my implementation, I use the mean of MJD-START and MJD-END, 
                but so long as ``mjd_obs`` is properly ordered (e.g. first observation to last observation) the use of MJD-START, MJD-END, or the mean is 
                largely irrelevant. However, it is necessary to have units of MJD (modified julian date).
            ra: list or array
                List or 1d array of the RA value in the spectra's ``.fits`` files. Should be ordered in accordance with ``mjd_obs``, e.g. ``mjd_obs[0]``
                corresponds to ``ra[0]`` and so on.
            dec: list or array
                List or 1d array of the DEC value in the spectra's ``.fits`` files. Should be ordered in accordance with ``mjd_obs``, e.g. ``mjd_obs[0]``
                corresponds to ``dec[0]`` and so on.
            location: EarthLocation
                ``astropy`` EarthLocation of the observation facility. Refer to the tutorial and ``astropy`` documentation for more information. The key 
                is that this value fullfills the ``location`` argument in ``astropy.time.Time``.
            a: float
                Semi-major axis of the exoplanet in units of AU. This can be found with NASA's Exoplanet Archive.
            P_orb: float
                Orbital period of the exoplanet in units of days. This can be found with NASA's Exoplanet Archive.
            i: float
                The inclination of the exoplanet in units of degrees. This can be found with NASA's Exoplanet Archive.
            T_not: float
                A mid-transit time of the exoplanet in units of modified julian days (MJD). It is not neccesary that 
                this date correponds to the specific transit in your data (I think). TBH I will have to double check 
                with some tests, and this value will likely be unnecessary in later iterations due to the ``transit_start_end``
                input. 
            v_sys: float
                The radial velocity of the exoplanetary system in units of km/s. This can be found with NASA's Exoplanet Archive.
            transit_start_end: tuple
                The start and end times of the transit in your data in units of modified julian date (MJD). Takes the form (start, end).
            gap_size: int
                Minimum number of NaNs between detector segments in your flux array to qualify as a separate detector. For CRIRES+, a good 
                value is ``gap_size=5``. However, test yourself to ensure all detectors are identified. You can do this by comparing the number of 
                detected detectors to those shown in, e.g. the ESO exposure time calculator https://www.eso.org/observing/etc/.
            remove_segments: array of ints
                Indices of the detectors that you do not want to include in the analysis. This is typically because some detectors are saturated 
                with telluric noise. E.g., in the CRIRES+ analysis, I did not want detectors 1, 2, 3, 4, 6, and 20; therefore ``remove_segments=[0, 1, 2, 3, 5, 19]``.
            first_components: int
                Number of components to remove starting from the first component found in PCA. The first components correspond to correlated signal, primarily the 
                stellar and telluric spectrum. It helps for this value to immediately disregard the first few components for this reason. 3-5 have 
                been good values in the past. If this value is too low, the pipeline could get trapped in a local maximum in the signal to noise (S/N), which 
                would lead to a suboptimal decomposition.
            last_components: int
                Number of components to remove starting from the last component found in PCA. The last components correspond to uncorrelated signal, e.g. instrumental 
                noise. It is less necessary to immediately disregard components here, particulary because the planetary signal sits right at the edge of the noise. 
                In the past, values from 0-3 have worked OK for me.
    Returns
    -------
    list
        Time-domain PCA reduction. Preferred over the wavelength domain. Takes the form of a list with the number of unremoved detectors as its length. Each index of the
        list contains a 2d array of shape n spectra x wavelength grid. The wavelength grids are often not consistent between detectors.
    list
        Wavelength-domain PCA reduction. Takes the form of a list with the number of unremoved detectors as its length. Each index of the list contains a 2d array of shape n spectra x wavelength grid. 
        The wavelength grids are often not consistent between detectors.
    list
        List of 1d wavelength grids for each unremoved detector.
    array
        2d CCF array (n spectra x velocity grid defined in ``v_shift_range``) in the rest frame of the Earth.
    array
        2d CCF array (n spectra x ``planet_frame_vgrid``) doppler corrected into the rest frame of the exoplanet. 
    array
        1d array of the velocity range for the exoplanetary CCF grid. Same format as ``v_shift_range``, however covers a smaller range.
    array
        2d CCF array (n in-transit spectra x ``planet_frame_vgrid``) of only in-transit spectra in the exoplanetary rest frame.
    array
        2d CCF array sampling values in a radial velocity amplitude array (``Kp_range`` x ``planet_frame_vgrid``). Refer to ccf_tests.sn_map.
    array
        2d signal to noise (S/N) map array (``Kp_range`` x ``planet_frame_vgrid``). Refer to ccf_tests.sn_map.
    array
        In-trail values of the S/N map. Refer to ccf_tests.welch_t_test.
    array
        Out-trail values of the S/N map. Refer to ccf_tests.welch_t_test.
    array
        T-statistic of the S/N map. Refer to ccf_tests.welch_t_test.
    array
        p value of the S/N map. Refer to ccf_tests.welch_t_test."""
    wave, flux, mjd_obs, ra, dec, location = kwargs['wave'], kwargs['flux'], kwargs['mjd_obs'], kwargs['ra'], kwargs['dec'], kwargs['location']
    a, P_orb, i, T_not, v_sys, transit_start_end = kwargs['a'], kwargs['P_orb'], kwargs['i'], kwargs['T_not'], kwargs['v_sys'], kwargs['transit_start_end']
    gap_size, remove_segments, first_components, last_components = kwargs['gap_size'], kwargs['remove_segments'], kwargs['first_components'], kwargs['last_components']

    debug_print(verbose, "Running pipeline...")
    debug_print(verbose, "Normalizing flux array...")
    normalized_flux_array, segment_indices = split_detectors(wave, flux, m=gap_size)

    if remove_segments is None:
        remove_segments = []

    # Filter segments
    keep_indices = [i for i in range(len(segment_indices)) if i not in remove_segments]
    debug_print(verbose, f"Retaining detector indices {keep_indices}")

    debug_print(verbose, "Running PCA subtraction on detector segments...")
    all_tdm, all_wdm, all_wave = [], [], []

    for keep_index in keep_indices:
        start, end = segment_indices[keep_index]
        #print("start, end: ", start, end)
        wave_i = wave[0, start:end]
        flux_i = normalized_flux_array[:, start:end]
        nanmask = ~np.isnan(wave_i) & ~np.isnan(flux_i[0])
        #print(flux_i[:, nanmask].shape)
        tdm_concat, wdm_concat = pca_subtraction(flux_i[:, nanmask], 0, np.sum(nanmask), first_comps=first_components, last_comps=last_components, pre=True, eighcalc=eighcalc)
        
        all_tdm.append(tdm_concat)
        all_wdm.append(wdm_concat)
        all_wave.append(wave_i[nanmask])    

    debug_print(verbose, "length of all_tdm: ", len(all_tdm))
    debug_print(verbose, "length of all_wdm: ", len(all_wdm))
    debug_print(verbose, "length of all_wave: ", len(all_wave))
    
    all_tdm = [np.array(x) for x in all_tdm]
    all_wdm = [np.array(x) for x in all_wdm]

    debug_print(verbose, "Running CCF on detector segments...")
    earth_frame_ccf, planet_frame_ccf, planet_frame_vgrid, in_transit = run_ccf_on_detector_segments(all_wave, 
                                 all_tdm, v_shift_range, keep_indices, sim_wave, 
                                 sim_flux, mjd_obs, ra, dec, location, 
                                 a, P_orb, i, T_not, v_sys, transit_start_end, verbose=verbose)
    
    debug_print(verbose, "Making the S/N map...")
    Kp_range_ccf, sn_map_array = sn_map(planet_frame_ccf, planet_frame_vgrid, Kp_range=Kp_range, **kwargs) 

    debug_print(verbose, "Performing Welch's t-test...")
    in_trail_vals, out_of_trail_vals, t_stat, p_value = welch_t_test(Kp_range_ccf)   
    
    debug_print(verbose, "Pipeline completed successfully.")
    return all_tdm, all_wdm, all_wave, earth_frame_ccf, planet_frame_ccf, planet_frame_vgrid, in_transit, Kp_range_ccf, sn_map_array, in_trail_vals, out_of_trail_vals, t_stat, p_value

def sample_components(start_components, stable_components, sim_wave, 
                      sim_flux, v_shift_range=np.linspace(-100_000, 100_000, 201), sn_test=-50, sn_max=-100, sample_end=False, results=None, verbose=True, **kwargs):
    """Samples through a range of the PCA component space to maximize S/N.
    
    Parameters
    ----------
    start_components: int
        Components to immediately remove from start or end. 
    stable_components: int
        Components of either the start or end that remain consistent. E.g. if the optimal first components have 
        already been found, ``stable_components`` is used to maintain the ``first_components`` value.
    sim_wave: array
        Refer to pipeline.pipeline.
    sim_flux: array
        Refer to pipeline.pipline. 
    sn_test: int
        Refer to pipeline.sample_full_pca_components.
    sn_max: int
        Refer to pipeline.sample_full_pca_components.
    sample_end: boolean
        True if optimizing for the best ``last_components``, False if sampling for the est ``first_components``.
    results: list
        Returned list of values from pipeline.pipeline.
    verbose: boolean
        Refer to untility_functions.debug_print.
    **kwargs
        Refer to pipeline.pipeline.
        
    Returns
    -------
    list
        List of optimized results from pipeline.pipeline."""
    # a, P_orb, i = kwargs['a'], kwargs['P_orb'], kwargs['i']

    while sn_test >= sn_max:

        best_results = results
        sn_max = sn_test
        best_components = start_components

        # Run pipeline
        if sample_end: 
            debug_print(verbose, f"sampling from the end. new sn_max = {sn_test} fc = {stable_components} lc = {start_components}")
            
            kwargs['first_components'] = stable_components 
            kwargs['last_components'] = start_components

            results = pipeline(
                sim_wave, sim_flux, v_shift_range=v_shift_range, verbose=verbose, **kwargs
            )

        else: 
            debug_print(verbose, f"sampling from the start. new sn_max = {sn_test} fc = {start_components} lc = {stable_components}")

            kwargs['first_components'] = start_components
            kwargs['last_components'] = stable_components

            results = pipeline(
                sim_wave, sim_flux, v_shift_range=v_shift_range, verbose=verbose, **kwargs
            )

        # Compute S/N
        sn_test = find_max_sn_in_expected_range(results[8], results[5] / 1000, **kwargs)
        debug_print(verbose, "sn_test =", sn_test)

        start_components += 1

    return best_results, sn_max, best_components    


def sample_full_pca_components(sim_wave, 
        sim_flux, v_shift_range=np.linspace(-100_000, 100_000, 201), sn_test=-50, sn_max=-100, verbose=True, **kwargs):
    """Loops pipeline() through the component space of the principal 
    component analysis, progressively removing the first components 
     (associated with the stellar spectrum and tellurics) until S/N in 
     the range of the exoplanetary parameters (+- 15 km/s from Kp=Kp, velocity in planet frame = 0)
    is maximized, then doing the same to the end components 
    (associated with uncorrelated/instrumental noise).
    
    Parameters
    ----------
    sim_wave: array
        Refer to pipeline.pipeline.
    sim_flux: array 
        Refer to pipeline.pipeline.
    sn_test: int
        Starting value for the S/N test value. Must be greater than ``sn_max`` but still start quite low.
    sn_max: int
        Starting value for the S/N max value. Must be less than ``sn_test``.
    verbose: boolean
        Refer to pipeline.pipeline
    **kwargs
        Refer to pipeline.pipeline.

    Returns
    -------
    list
        List of the returned values in pipeline.pipeline for the optimal range of PCA components.
    int
        Optimal value for ``first_components``.
    int
        Optimal value for ``last_components``.
    float
        Maximum S/N value in planetary parameter range. 
    """
    first_components, last_components = kwargs['first_components'], kwargs['last_components']

    first_best_results, first_sn_max, first_best_components = sample_components(
        first_components, last_components, sim_wave, 
        sim_flux, v_shift_range=v_shift_range, sn_test=sn_test, sn_max=sn_max, sample_end=False, verbose=verbose, **kwargs)

    best_results, sn_max, last_best_components = sample_components(
        last_components, first_best_components - 1, sim_wave, 
        sim_flux, v_shift_range=v_shift_range, sn_test=first_sn_max, sn_max=sn_max, sample_end=True, results=first_best_results, verbose=verbose, **kwargs)

    debug_print(verbose, f"Best fc = {first_best_components - 1}, best lc = {last_best_components - 1}, S/N = {sn_max}")

    return best_results, first_best_components - 1, last_best_components - 1, sn_max
    



# Follows the steps of Damiano 2018, Sec. 5.3

"""
1. normalize each spectrum of each detector dividing by its median
2. take the mean spectrum of each detector
3. take all lines <0.8 & also in telluric template
4. fit each line to a gaussian, take the centroid
5. plot pixel position vs. wavelength
6. fit a polynomial to the centroid positions until no correlated residuals
7. find fit precision: deltaV (velocity) = std(residuals) * speed of light / central spectrum wavelength
8. interpolate all single spectra by a third order spline to derived wavelength grid (converting pixel to wavelength via the fit eqn)
"""

import numpy as np

from numpy.polynomial import Polynomial
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from scipy.signal import resample, find_peaks
from astropy.constants import c

from pcassie.utility_functions import split_divide_by_median

def gaussian(x, amp, mu, sigma, offset):
    """A simple gaussian function."""
    return amp * np.exp(-0.5 * ((x - mu) / sigma) ** 2) + offset


def fit_gaussian_to_peaks(x, y, peaks, window=5):
    """Fit peak absoption features to gaussians."""
    centroids = []
    for idx in peaks:
        left = max(0, idx - window)
        right = min(len(x), idx + window + 1)
        x_fit = x[left:right]
        y_fit = y[left:right]
        # Initial guesses
        amp_guess = np.min(y_fit) - np.median(y_fit)
        mu_guess = x[idx]
        sigma_guess = (x_fit[-1] - x_fit[0]) / 6 if len(x_fit) > 1 else 1
        offset_guess = np.median(y_fit)
        try:
            popt, _ = curve_fit(
                gaussian, x_fit, y_fit,
                p0=[amp_guess, mu_guess, sigma_guess, offset_guess]
            )
            centroids.append(popt[1])
        except Exception:
            centroids.append(mu_guess)
    return np.array(centroids)


def fit_segments_to_wavelengths(segment_centroids, tel_centroid_dict, telluric_wavelength, deg=3):
    """Fits individual detectors to a new wavelength grid."""
    fits = {}

    pixel_grid = np.arange(len(telluric_wavelength))
    interp_wave = interp1d(pixel_grid, telluric_wavelength, kind='linear', bounds_error=False, fill_value=np.nan)

    # Interpolator: wavelength to pixel index
    interp_idx = interp1d(telluric_wavelength, np.arange(len(telluric_wavelength)), 
                      kind='linear', bounds_error=False, fill_value=np.nan)

    tel_centroids_pixel = {seg_id: interp_idx(waves) for seg_id, waves in tel_centroid_dict.items()}

    for seg_id in segment_centroids:
        data_pix = segment_centroids[seg_id]
        tel_pix = tel_centroids_pixel[seg_id]

        # Skip if either segment is empty
        if len(data_pix) == 0 or len(tel_pix) == 0:
            continue

        # Interpolate wavelength at telluric centroids
        wave_vals = interp_wave(tel_pix)

        # Fit pixel → wavelength
        mask = ~np.isnan(data_pix) & ~np.isnan(wave_vals)
        p = Polynomial.fit(data_pix[mask], wave_vals[mask], deg=deg).convert()
        print(p)
        fits[seg_id] = p

    return fits


def precision(residuals, wave_arr):
    """A function to calculate the precision of the wavelength fit in terms of velocity. 
    Refernce https://discovery.ucl.ac.uk/id/eprint/10066066/7/Mario_Damiano_Thesis.pdf 
    pg. 117 for more info."""
    std = np.std(residuals)
    central_wave = np.median(wave_arr)
    return std * c / (central_wave)


def split_and_stack(arr, gaps):
    """Rearranges your spectrum into a 3d array with (detector, spectra, flux); 
    i.e. a shape of (n detector x n spectra x wavelength range)"""
    segments = []
    gaps = np.concatenate(([0], gaps))
    for i in range(len(gaps) - 1):
        segments.append(arr[gaps[i]:gaps[i + 1]].astype(float))  # cast to float
    segments.append(arr[gaps[-1]:].astype(float))
    maxlen = max(len(seg) for seg in segments)
    stacked = np.array([np.pad(seg, (0, maxlen - len(seg)), constant_values=np.nan) for seg in segments])
    return stacked


def calibrate_cr2res(data_wave, data_flux, telluric_wave, telluric_flux, gap_size_px=5, poly_order=5):
    """
    Runs a full calibration of your spectrum. 

    Args: 
    data_flux is an array of n spectra x wavelength array
    data_wave is a 1d wavelength array
    telluric_flux is a 1d telluric flux array
    telluric_wave is a 1d telluric wavelength array

    Ensure the wavelegnth units match :)
    """

    #1.1 mask data of NaNs
    pixels = np.arange(len(data_wave))
    valid_mask = ~np.isnan(data_flux).any(axis=0)  # Mask out nans

    valid_flux = data_flux[:, valid_mask]
    valid_pixels = pixels[valid_mask]
    valid_wave = data_wave[valid_mask]    

    print("flux, wavelength array shape", data_flux.shape, data_wave.shape)

    print("masked flux, pixel array shape", valid_flux.shape, valid_pixels.shape)

    # 1.2 divide each spectrum of each detector by its median
    normalized_flux_array = []
    gaps_arr = []  # Store gaps for each spectrum

    for ii in range(len(valid_flux)):
        orig_flux = valid_flux[ii, :]
        
        orig_norm_flux, gaps = split_divide_by_median(valid_wave, orig_flux, gap_size_px)
        
        normalized_flux_array.append(orig_norm_flux)
        gaps_arr.append(gaps)

    # 2 Take the mean spectrum of each detector
    mean_flux = np.nanmean(normalized_flux_array, axis=0)

    # 3 interpolate tellurics to ~data from the given wavelength range
    # Mask telluric arrays to this range
    data_wave_min = np.min(data_wave) 
    data_wave_max = np.max(data_wave)

    # Mask telluric arrays to this range
    mask = (telluric_wave >= data_wave_min) & (telluric_wave <= data_wave_max)
    telluric_wave_masked = telluric_wave[mask]
    telluric_flux_masked = telluric_flux[mask]

    # Resample telluric arrays to match the shape of the data arrays
    telluric_wave_resampled = resample(telluric_wave_masked, len(valid_pixels))
    telluric_flux_resampled = resample(telluric_flux_masked, len(valid_pixels))

    # 3.2 Isolate data peaks <0.8
    all_peaks, _ = find_peaks(-mean_flux)  # Find minima
    data_peaks = all_peaks[mean_flux[all_peaks] < 0.8]

    # 3.3 Isolate top-N deepest telluric peaks
    all_tel_peaks, _ = find_peaks(-telluric_flux_resampled)
    peak_fluxes = telluric_flux_resampled[all_tel_peaks]

    N = len(data_peaks)
    top_idx = np.argsort(peak_fluxes)[:N]              # deepest N
    tel_peaks_unsorted = all_tel_peaks[top_idx]

    # Sort by wavelength (assuming telluric_wavelength is defined)
    tel_wavelengths = telluric_wave_resampled[tel_peaks_unsorted]
    sorted_idx = np.argsort(tel_wavelengths)

    tel_peaks = tel_peaks_unsorted[sorted_idx]

    # 4 fit peaks to gaussians
    data_centroids = fit_gaussian_to_peaks(valid_pixels, mean_flux, data_peaks)
    tel_centroids = fit_gaussian_to_peaks(telluric_wave_resampled, telluric_flux_resampled, tel_peaks)

    # 5.1.1 Separate data_centroids by the gaps they fall into

    # data_centroids are pixel indices; use valid_pixels[data_centroids] if needed for mapping to pixel values
    # If you want to use the pixel values directly, use data_centroids as is

    sep_data_centroids = []
    gaps_with_ends = np.concatenate(([0], gaps, [valid_pixels[-1]]))  # prepend 0 for the first segment

    for i in range(len(gaps_with_ends) - 1):
        start = gaps_with_ends[i]
        end = gaps_with_ends[i + 1]
        # Select centroids that fall within this segment
        mask = (data_centroids >= start) & (data_centroids < end)
        sep_data_centroids.append(data_centroids[mask])

    # sep_data_centroids is a list of arrays, one per segment
    for idx, arr in enumerate(sep_data_centroids):
        print(f"Segment {idx}: {len(arr)} centroids")

    segment_centroids = {seg_id: arr for seg_id, arr in enumerate(sep_data_centroids)}

    # Get segment lengths from data_centroids
    segment_lengths = [len(arr) for arr in segment_centroids.values()]

    # Split telluric_centroids into matching segments
    split_tel_centroids = np.split(tel_centroids, np.cumsum(segment_lengths)[:-1])

    # Build dict with same keys as segment_centroids
    tel_centroid_dict = {i: seg for i, seg in enumerate(split_tel_centroids)}
        
    # Run the fit
    segment_fits = fit_segments_to_wavelengths(segment_centroids, tel_centroid_dict, telluric_wave_resampled, deg=poly_order)

    #6 get residuals
    residuals_dict = {}

    for seg_id, poly in segment_fits.items():
        pixels = np.atleast_1d(segment_centroids[seg_id])
        true_wavelengths = np.atleast_1d(tel_centroid_dict[seg_id])

        if len(pixels) != len(true_wavelengths):
            print(f"Skipping segment {seg_id}: mismatched lengths")
            continue

        # Evaluate fit
        fitted_wavelengths = poly(pixels)
        residuals = fitted_wavelengths - true_wavelengths
        residuals_dict[seg_id] = residuals

    #7. find fit precision

    segment_precision = {}

    for seg_id, residual in residuals_dict.items():
        
        pix = segment_centroids[seg_id]
        if len(pix) == 0:
            continue

        # Use full pixel span for this segment
        full_pix_range = np.arange(int(np.min(pix)), int(np.max(pix)) + 1)

        # Evaluate fitted poly at full pixel range
        fitted_waves = poly(full_pix_range)

        # Remove any nan pairs
        mask = ~np.isnan(fitted_waves) 

        # Compute precision
        prec = precision(residual, fitted_waves[mask])
        segment_precision[seg_id] = prec

    # 8 interpolate new wavelength grid
    sep_pixels = split_and_stack(valid_pixels, gaps_arr[0])
    data_wavelengths = np.full_like(mean_flux, np.nan, dtype=float)

    # Convert each segment's pixels to wavelength using fitted polynomials
    for seg_id, poly in segment_fits.items():
        pixels = sep_pixels[seg_id]
        pixels = pixels.astype(int)

        # Ensure we don't go out of bounds
        pixels = pixels[(pixels >= 0) & (pixels < len(mean_flux))]
        data_wavelengths[pixels] = poly(pixels)

    return data_wavelengths


import numpy as np

from scipy.stats import ttest_ind
from scipy.interpolate import interp1d

from pcassie.ccf import *

def inject_simulated_signal(sim_wave, sim_flux,  
                            R_p, R_star, multiple=1, verbose=False, **kwargs):
    """
    Inject a simulated signal into the observed flux array. Both flux and sim_flux must be normalized.

    Parameters
    ----------
    sim_wave: array
        Refer to pipeline.pipeline.
    sim_flux: array
        Refer to pipeline.pipeline.
    R_p: float
        Radius of the exoplanet. Units must be the same as R_star.
    R_star: float
        Radius of the host star. Units must be tha same as R_p.
    multiple: int
        Scalar multiple of the injection. E.g. if you want the injected signal to be 0.5x the scale of the 
        expected planet signal, set ``multiple=0.5``.
    verbose: boolean
        Refer to utility_functions.debug_print.
    **kwargs
        Refer to pipeline.pipeline.

    Returns
    -------
    array
        new flux array with injected signal
    """
    wave, flux, mjd_obs, ra, dec, location = kwargs['wave'][0], kwargs['flux'], kwargs['mjd_obs'], kwargs['ra'], kwargs['dec'], kwargs['location']
    a, P_orb, i, T_not, v_sys = kwargs['a'], kwargs['P_orb'], kwargs['i'], kwargs['T_not'], kwargs['v_sys']

    factor = (R_p / R_star) * multiple
    v_bary = compute_vbary_timeseries(ra, dec, mjd_obs, location)
    correction = doppler_correction(a=a, P_orb=P_orb, i=i, t=mjd_obs, T_not=T_not, v_sys=v_sys, v_bary=v_bary)
    debug_print(verbose, f"correction: {correction}")
    sim_on_obs_grid = interp1d(sim_wave, sim_flux, bounds_error=False, fill_value=0)
    spectra_grid = np.zeros_like(flux)

    sim_shifts = []
    for j in range(len(mjd_obs)):
        shifted_wave = doppler_shift(wave, correction[j])  # ← shift wavelengths, not flux
        shifted_sim = sim_on_obs_grid(shifted_wave) * factor
        spectra_grid[j, :] = flux[j, :] + shifted_sim
        sim_shifts.append(shifted_sim)
    return spectra_grid

def sn_map(
    planet_frame_ccf, planet_frame_vgrid,
    Kp_range=np.linspace(50_000, 150_000, 101), **kwargs
):
    """
    SNR map computation for the CCF.

    Parameters
    ----------
    planet_frame_ccf: array
        2d CCF array in the planet rest frame. Refer to pipeline.pipeline.
    planet_frame_vgrid: array
        1d velocity grid (km/s). Refer to pipeline.pipeline.
    Kp_range: array
        Range of velocities in m/s to sample radial velocity amplitude Kp. Like ``v_shift_range`` and ``planet_frame_vgrid``, ideal to have in steps of 1 km/s.
    **kwargs
        Refer to pipeline.pipeline.

    Returns
    -------
    array
        2d grid of CCF values over the sampled radial velocity amplitude (Kp) vs. velocity shift space.
    array
        2d S/N grid made by dividing the Kp vs. velocity shift grid by its the standard deviation of the out trail values.
    """
    mjd_obs, ra, dec, location = kwargs['mjd_obs'], kwargs['ra'], kwargs['dec'], kwargs['location']
    a, P_orb, i, T_not, v_sys, transit_start_end = kwargs['a'], kwargs['P_orb'], kwargs['i'], kwargs['T_not'], kwargs['v_sys'], kwargs['transit_start_end']

    n_Kp = len(Kp_range)
    n_v = len(planet_frame_vgrid)

    # Preallocate result array instead of appending
    Kp_range_ccf = np.zeros((n_Kp, n_v), dtype=np.float32)
    
    for idx, Kp in enumerate(Kp_range):
        try:
            # Doppler correct
            this_cropped_ccf, _ = doppler_correct_ccf(
                planet_frame_ccf, planet_frame_vgrid, mjd_obs,
                ra, dec, location, a, P_orb, i, T_not, v_sys, Kp=Kp
            )

            if np.isnan(this_cropped_ccf).any():
                continue  # Skip Kp values with NaNs

            # Remove out-of-transit
            ccf_in_transit = remove_out_of_transit(
                transit_start_end=transit_start_end,
                grid=this_cropped_ccf,
                mjd_obs=mjd_obs
            )

            # Sum across time axis (axis=0)
            Kp_range_ccf[idx] = np.sum(ccf_in_transit, axis=0)

        except Exception as e:
            print(f"Skipping Kp = {Kp:.1f} due to error: {e}")
            continue

    # Mask near-planet velocities (±15 km/s = 15000 m/s)
    exclude_planet_mask = (np.abs(planet_frame_vgrid) < 15000)

    # Standard deviation of CCF away from planet signal
    outside_std = np.std(Kp_range_ccf[:, exclude_planet_mask], axis=1)

    # Avoid division by zero
    outside_std[outside_std == 0] = np.nan

    # Compute S/N map
    sn_map_array = Kp_range_ccf / outside_std[:, np.newaxis]

    return Kp_range_ccf, sn_map_array


def welch_t_test(Kp_range_ccf, zoom_radius=15):
    """Performs Welch's T-test to compare values in the CCF grid
    that may be associated with the planet with those that are not.
    
    Parameters
    ----------
    Kp_range_ccf: array
        2d Kp grid from ccf_tests.sn_map. Refer to ccf_tests.sn_map.
    zoom_radius: int, optional
        Pixel space radius to build the in trail box around. Here, 1 pixel = 1 km/s (if you have your units right).

    Returns
    -------
    array
        2d Kp vs. velocity shift CCF grid for in trail values.
    array
        2d Kp vs. velocity shift CCF grid for out trail values.
    int
        Welch's T-test statistic comparing in trail and out trail CCF values.
    int
        Welch's p value comparing in trail and out trail CCF values.
    """
    # Define zoom radius in pixels (typically km/s)
    Kp_range_ccf = np.array(Kp_range_ccf)  # Ensure it's a NumPy array

    # Step 1: Find max index in S/N map
    max_val = np.nanmax(Kp_range_ccf)
    max_idx = np.argwhere(Kp_range_ccf == max_val)[0]
    max_row, max_col = max_idx

    # Step 2: Clip the box to array boundaries
    n_rows, n_cols = Kp_range_ccf.shape
    min_row = max(0, max_row - zoom_radius)
    max_row_clip = min(n_rows, max_row + zoom_radius + 1)  # +1 because slicing is exclusive
    min_col = max(0, max_col - zoom_radius)
    max_col_clip = min(n_cols, max_col + zoom_radius + 1)

    # Step 3: Extract in-trail values
    in_trail_vals = Kp_range_ccf[min_row:max_row_clip, min_col:max_col_clip].ravel()

    # Step 4: Create out-of-trail values by masking in-trail region
    masked_array = Kp_range_ccf.copy()
    masked_array[min_row:max_row_clip, min_col:max_col_clip] = np.nan
    out_of_trail_vals = masked_array[~np.isnan(masked_array)]

    # Step 5: Welch’s t-test
    t_stat, p_value = ttest_ind(in_trail_vals, out_of_trail_vals, equal_var=False)

    return in_trail_vals, out_of_trail_vals, t_stat, p_value


def find_max_sn_in_expected_range(sn_array, v_grid, offset=75, zoom_radius=15, **kwargs):
    """Finds the maximum S/N value in the range of values expected of the planet.
    
    Parameters
    ----------
    
    sn_array: array
        2d S/N map from ccf_tests.sn_map.
    v_grid: array
        1d velocity range corresponding to ``sn_array``.
    offset: int, optional
        Valocity offset of the Kp range. E.g. if you were to sample a CCF from 75 km/s < Kp < 175 km/s, set ``offset=75``.
    zoom_radius: int
        Refer to ccf_test.welch_t_test.
    **kwargs
        Refer to pipeline.pipeline.

    Returns
    -------
    int
        Maximum S/N value within the ``zoom_radius`` of the expected Kp and velocity shift values (expected v shift value in 
        planet frame is 0).
    """

    a, P_orb, i = kwargs['a'], kwargs['P_orb'], kwargs['i']
    Kp = rv_amplitude(a * 1.495979e11, P_orb * 24 * 3600, np.radians(i)) / 1000
    #print(Kp)

    row_idx = int(Kp) - offset
    col_idx = np.argwhere(v_grid == 0)[0][0]

    #print(row_idx, col_idx)

    n_rows, n_cols = sn_array.shape
    min_row = max(0, row_idx - zoom_radius)
    max_row_clip = min(n_rows, row_idx + zoom_radius + 1)  # +1 because slicing is exclusive
    min_col = max(0, col_idx - zoom_radius)
    max_col_clip = min(n_cols, col_idx + zoom_radius + 1)
    #print(min_row, max_row_clip, min_col, max_col_clip)

    expected_range = sn_array[min_row:max_row_clip, min_col:max_col_clip]

    return np.max(expected_range)  


import setuptools

with open("README.md", "r") as fh: 
    long_description = fh.read()

setuptools.setup(
    name="pcassie",
    version="0.1.7",
    author="Kenny T. Phan",
    author_email="kenny.phan@yale.edu",
    description="A Pincipal Component Pipeline based on Damiano et al. 2019",
    packages=["pcassie"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: MIT License",   
        "Intended Audience :: Science/Research", 
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Physics"
    ]
)

install_requires = ["numpy", "matplotlib.pyplot", "scipy", "jax", "pandas", "numba", "astropy"]




import numpy as np
from numba import njit
import os
os.environ["JAX_PLATFORM_NAME"] = "cpu"
import jax.numpy as jnp
from jax import config
config.update("jax_enable_x64", True)

from pcassie.utility_functions import debug_print

def convert_range_to_indices(wave, start, end):
    """Convert a wavelength range to indices.
    
    Parameters
    ----------
    wave: array
        1d wavelength grid.
    start: float
        Starting wavelength value you want to crop to.
    end: float
        Ending wavelength value you want to crop to.

    Returns:
    --------
    int
        Index of the starting wavelength.
    int
        Index of the ending wavelength.
    """
    start_index = np.searchsorted(wave, start)
    end_index = np.searchsorted(wave, end)
    return start_index, end_index

# def forward_preprocess(spectra):
#     """
#     Returns:
#       preprocessed (n_rows, n_cols),
#       row_std (n_rows, 1),
#       col_med (1, n_cols)  # keep as 2D for easy broadcasting
#       global_med (scalar)
#     """
#     global_med = np.median(spectra)                     # scalar
#     norm_flux = spectra / global_med                    # shape (n_rows, n_cols)
#     col_med = np.median(norm_flux, axis=0, keepdims=True)  # shape (1, n_cols)
#     median_subtracted = norm_flux - col_med             # shape (n_rows, n_cols)

#     row_std = np.linalg.norm(median_subtracted, axis=1, keepdims=True) \
#               / np.sqrt(median_subtracted.shape[1])
#     row_std = np.where(row_std == 0.0, 1.0, row_std)    # avoid zeros

#     preprocessed = median_subtracted / row_std          # same shape (n_rows, n_cols)
#     return preprocessed, row_std, col_med, global_med


# def inverse_preprocess(preprocessed_slice, row_std, col_med, global_med, start_idx, end_idx):
#     """
#     Invert preprocessing for just the wavelength slice [start_idx:end_idx).
#     preprocessed_slice: shape (n_rows, n_slice_cols)
#     row_std: shape (n_rows, 1)
#     col_med: shape (1, n_cols_total)  <-- we'll index into the relevant slice
#     global_med: scalar
#     start_idx, end_idx: slice indices to pick the right columns from col_med
#     """
#     # 1) undo row normalization
#     median_subtracted = preprocessed_slice * row_std                     # broadcasts along columns

#     # 2) add back column median for that slice
#     col_med_slice = col_med[:, start_idx:end_idx]                        # shape (1, n_slice_cols)
#     norm_flux_slice = median_subtracted + col_med_slice                  # shapes broadcast

#     # 3) undo global normalization
#     reconstructed_slice = norm_flux_slice * global_med                   # shape (n_rows, n_slice_cols)
#     return reconstructed_slice

def preprocess(spectra):
    """
    Normalize by the median spectrum, subtract the median at each wavelength,
    and divide each spectrum by its own standard deviation.

    Parameters
    ----------
    spectra: array
        2d spectral flux grid.

    Returns
    -------
    array
        Median subtracted, standard deviation divided 2d spectral flux grid.
    """
    norm_flux = spectra / np.median(spectra)
    median_flux = np.median(norm_flux, axis=0)
    median_subtracted = norm_flux - median_flux

    row_std = np.linalg.norm(median_subtracted, axis=1, keepdims=True) \
              / np.sqrt(median_subtracted.shape[1])

    # avoid division by zero
    row_std = np.where(row_std == 0.0, 1.0, row_std)

    return median_subtracted / row_std, row_std

def compute_covariance_matrix(data):
    """Compute the covariance matrix using NumPy (faster than pandas).
    
    Parameters
    ----------
    data: array
        2d spectral array. Used after PCA analysis to the Time Domain or Wavelength Domain.
    
    Returns
    -------
    array
        Covariance matrix."""
    centered = data - np.mean(data, axis=0)
    return centered.T @ centered / (data.shape[0] - 1)


def compute_eigenvalues_and_vectors_jax(cov_matrix):
    """Compute and sort eigenvalues/eigenvectors in descending order.
    
    Parameters
    ----------
    cov_matrix: array
        Covariance matrix. Refer to pca_subtraction.compute_covariance_matrix.
        
    Returns
    -------
    array
        1d array of the eigenvalues in the order from highest to lowest (I think).
    array
        2d array of eigenvectors in the order of their corresponding eiganvalues."""
    jax_cov_matrix = jnp.array(cov_matrix, dtype=jnp.float64)
    evals, evecs = jnp.linalg.eigh(jax_cov_matrix)
    idx = jnp.argsort(evals)[::-1]

    evals_sorted = np.array(evals[idx])
    evecs_sorted = np.array(evecs[:, idx])

    return evals_sorted, evecs_sorted

@njit
def compute_eigenvalues_and_vectors_numba(cov_matrix):
    evals, evecs = np.linalg.eigh(cov_matrix)
    idx = np.argsort(evals)[::-1]   
    return evals, evecs, idx 


def explained_variance(eigenvalues):
    """Calculate explained variance ratio.
    
    Parameters
    ----------
    eigenvalues: array
        1d array of eigenvalues.
        
    Returns
    -------
    array
        Explained variance value for each eigenvalue."""
    return eigenvalues / np.sum(eigenvalues)


def remove_components(data, eigenvectors, first_comps=0, last_comps=0, verbose=False):
    """Remove specified principal components from the data.
    
    Parameters:
    data: array
        2d flux array.
    eigenvectors: array
        2d eigenvectors. Refer to ``pca_subtraction.compute_eigenvalues_and_vectors_jax``.
    first_comps: int, optional
        Index of first components (eigenvectors) to remove.
    last_comps: int, optional
        Index of last components (eigenvectors) to remove.
    verbose: boolean
        Refer to ``utility_functions.debug_print``.
        
    Returns
    -------
    array
        2d flux array after removing the ``first_comps`` and ``last_comps``."""
    total_comps = eigenvectors.shape[1]
    start_comps = first_comps
    end_comps = total_comps - last_comps

    if start_comps >= end_comps:
        debug_print(verbose, f"total # of components: {total_comps}. removing {start_comps} from the beginning and {last_comps} from the end")
        raise ValueError("Requested to remove all components — nothing left to reconstruct from.")

    proj_matrix = eigenvectors[:, start_comps:end_comps]
    projected = data @ proj_matrix
    return projected @ proj_matrix.T


def pca_subtraction(spectra, start_idx, end_idx, first_comps=0, last_comps=0, eighcalc='numba', pre=False, verbose=False):
    """
    Perform PCA subtraction in a wavelength slice from `start_idx` to `end_idx`.

    Parameters
    ----------
    spectra (np.ndarray): 2D array of shape (num_spectra, num_wavelengths).
    start_idx (int): Start index for PCA region.
    end_idx (int): End index for PCA region.
    first_comps (int): Components to remove from the beginning.
    last_comps (int): Components to remove from the end.
    pre (bool): Whether to apply preprocessing first.

    Returns:
        (tdm_result, wdm_result): PCA-subtracted arrays.
    """
    if pre:
        spectra_pre, _ = preprocess(spectra)
    else:
        spectra_pre = spectra

    spectra_slice = spectra_pre[:, start_idx:end_idx]
    tdm = spectra_slice.T  # Transpose for TDM
    wdm = spectra_slice     # WDM as-is

    tdm_cov = compute_covariance_matrix(tdm) # spectra x spectra
    wdm_cov = compute_covariance_matrix(wdm) # wave x wave

    if eighcalc == 'jax':
        _, evec_tdm = compute_eigenvalues_and_vectors_jax(tdm_cov)
        _, evec_wdm = compute_eigenvalues_and_vectors_jax(wdm_cov)

    elif eighcalc == 'numba':
        _, evec_tdm, idx_tdm = compute_eigenvalues_and_vectors_numba(tdm_cov)
        evec_tdm = np.array(evec_tdm[:, idx_tdm])

        _, evec_wdm, idx_wdm  = compute_eigenvalues_and_vectors_numba(wdm_cov)
        evec_wdm = np.array(evec_wdm[:, idx_wdm])

    debug_print(verbose, "tdm, wdm evec shapes:", evec_tdm.shape, evec_wdm.shape)

    # PCA removal
    tdm_clean = remove_components(tdm, evec_tdm, first_comps, last_comps).T
    wdm_clean = remove_components(wdm, evec_wdm, first_comps, last_comps)
    # need to divide each column by std of the column

    return tdm_clean, wdm_clean